// cl/algorithms/cmean.cpp
#include<cl/algorithms/cmean.hpp>
#include<cl/errors.hpp>
#include<cl/distances/euclideandistance.hpp>
#include<iostream> 
#include<cmath> 
#include<boost/random.hpp>

namespace ClusLib {

    void Cmean::performClustering() const { 
        initialization(); 
        iteration(); 
    }

    void Cmean::setupArguments() {
        Algorithm::setupArguments();
        ASSERT(_ds->is_numeric(), "not a numeric dataset");

        _epsilon = boost::any_cast<Real>(
            _arguments.get("epsilon")); 
        ASSERT(_epsilon>0, "epsilon must be positive");

        _threshold = boost::any_cast<Real>(
            _arguments.get("threshold"));
        ASSERT(_threshold>EPSILON, "invalid threshold");

        _alpha = boost::any_cast<Real>(_arguments.get("alpha"));
        ASSERT(_alpha>1, "invalid alpha");

        _numclust = boost::any_cast<Size>(
            _arguments.get("numclust")); 
        ASSERT(_numclust>=2 && _numclust<=_ds->size(), 
            "invalid numclust");

        _maxiter = boost::any_cast<Size>(
            _arguments.get("maxiter")); 
        ASSERT(_maxiter>0, "invalide maxiter");

        _seed = boost::any_cast<Size>(
            _arguments.get("seed")); 
        ASSERT(_seed>0, "invalide seed");

        _distance = boost::shared_ptr<Distance>(new 
            EuclideanDistance());
    }

    void Cmean::fetchResults() const {
        Size s;
        for(Size i=0;i<_ds->size();++i) {
            Real dMax = MIN_REAL;
            for(Size k=0;k<_numclust;++k) { 
                if (dMax <  _FCM(i,k) ) {
                    dMax =  _FCM(i,k);
                    s = k;
                }
            }
            _CM[i] = s;
            _clusters[s]->add((*_ds)[i]);
        } 

        PClustering pc;
        for(Size i=0;i<_clusters.size();++i){
            pc.add(_clusters[i]);
        }
        _results.CM = _CM;
        _results.insert("pc",pc);
        _results.insert("fcm", _FCM);
        _results.insert("numiter", _numiter);
        _results.insert("dObj", _dObj);
    }

    void Cmean::iteration() const { 
        _numiter = 0;
        Real dPrevObj;
        while(true) { 
            updateCenter();
            updateFCM();

            dPrevObj = _dObj;
            calculateObj();

            ++_numiter;

            if (std::fabs(_dObj - dPrevObj) < _threshold){
                break;
            }

            if (_numiter >= _maxiter){
                break;
            }
        } 
    }

    void Cmean::updateCenter() const {
        Real dSum1, dSum2, dTemp;
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size k=0;k<_numclust;++k){ 
            for(Size j=0;j<schema->size();++j){
                dSum1 = 0.0;
                dSum2 = 0.0;
                for(Size i=0; i<_ds->size();++i){
                    dTemp = std::pow(_FCM(i,k),_alpha);
                    dSum1 += dTemp * 
                        (*schema)[j]->get_c_val((*_ds)(i,j));
                    dSum2 += dTemp;
                } 
                (*schema)[j]->set_c_val(
                    (*_clusters[k]->center())[j],dSum1/dSum2);
            }
        }
    }

    void Cmean::updateFCM() const {
        Real dSum, dTemp;
        std::vector<Real> dvTemp(_numclust);
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size i=0;i<_ds->size();++i){
            dSum = 0.0;
            for(Size k=0;k<_numclust;++k){
                dTemp = (*_distance)((*_ds)[i], 
                    _clusters[k]->center()) + _epsilon;
                dvTemp[k] = std::pow(dTemp, 2/(_alpha-1));
                dSum += 1 / dvTemp[k];
            }
            for(Size k=0;k<_numclust;++k){
                _FCM(i,k) = 1.0 / (dvTemp[k] * dSum);
            }
        }
    }

    void Cmean::initialization() const {
        Size numRecords = _ds->size();
        _FCM.resize(numRecords, _numclust);
        _CM.resize(numRecords, Null<Size>());

        std::vector<Size> index(numRecords,0);
        for(Size i=0;i<index.size();++i){
            index[i] = i;
        }

        boost::minstd_rand generator(_seed);
        for(Size i=0;i<_numclust;++i){
            boost::uniform_int<> uni_dist(0,numRecords-i-1);
            boost::variate_generator<boost::minstd_rand&, 
                boost::uniform_int<> > uni(generator,uni_dist); 
            Size r = uni();
            boost::shared_ptr<Record> cr = 
                boost::shared_ptr<Record>(new Record(*(*_ds)[r]));
            boost::shared_ptr<CenterCluster> c = 
                boost::shared_ptr<CenterCluster>(
                    new CenterCluster(cr)); 
            c->set_id(i);
            _clusters.push_back(c);
            index.erase(index.begin()+r);
        } 

        updateFCM();
    } 

    void Cmean::calculateObj() const {
        Real dSum = 0.0; 
        for(Size i=0; i<_ds->size(); ++i) {
            for(Size j=0; j<_numclust; ++j) {
                Real dTemp = (*_distance)((*_ds)[i], 
                    _clusters[j]->center());
                dSum += std::pow(_FCM(i, j), _alpha) * 
                    std::pow(dTemp, 2.0);
            }
        }
        _dObj = dSum;
    }
}
