// cl/algorithms/diana.cpp
#include<cl/algorithms/diana.hpp>
#include<cl/patterns/pcvisitor.hpp>
#include<cl/types.hpp>
#include<iostream>
#include<boost/pointer_cast.hpp>

namespace ClusLib {

    void Diana::setupArguments() {
        Algorithm::setupArguments();

        _distance = _arguments.distance;
        ASSERT(_distance, "distance is null");
    }

    void Diana::create_dm() const { 
        Size n = _ds->size();
        for(Size i=0;i<n-1;++i){
            for(Size j=i+1;j<n;++j){
                _dm.add_item(i, j, 
                      (*_distance)((*_ds)[i],(*_ds)[j]));
            }
        }
    }

    void Diana::performClustering() const {
        create_dm();
        initialization();
        division();
    }

    void Diana::initialization() const {
        Size n = _ds->size(); 
        Size id = 2*n-2;
        boost::shared_ptr<InternalNode> pin(new InternalNode(id));
        for(Size s=0;s<n;++s){
            boost::shared_ptr<LeafNode> pln(new 
                LeafNode((*_ds)[s], s));
            pln->set_level(0);
            pin->add(pln);
            _leaf.insert(std::pair<Size, 
                boost::shared_ptr<LeafNode> >(s, pln));
        }
        _internal.insert(std::pair<Size, 
            boost::shared_ptr<InternalNode> >(id, pin));
        _unsplitClusters.insert(id);

        Real dMax = MIN_REAL;
        for(Size i=0;i<n-1;++i){
            for(Size j=i+1;j<n;++j){
                if (dMax < _dm(i,j) ) {
                    dMax = _dm(i,j);
                }
            }
        }
        _clusterDiameter.insert(std::pair<Size, Real>(id,dMax));

        for(Size s=2*n-3; s>n-1; --s) {
            _clusterID.insert(s);
        } 
    }

    void Diana::division() const {
        Size n = _ds->size();
        std::set<Size>::iterator it;
        Real dMax;
        Size ind;
        for(Size s=2*n-2; s>n-1; --s) {
            dMax= MIN_REAL;
            std::vector<Size> nvTemp(_unsplitClusters.begin(),
                    _unsplitClusters.end());
            for(Size i=0; i<nvTemp.size(); ++i) {
                if(dMax < _clusterDiameter[nvTemp[i]]) {
                    dMax = _clusterDiameter[nvTemp[i]];
                    ind = nvTemp[i];
                }
            }

            _internal[ind]->set_level(s-n+1);
            _internal[ind]->set_id(s);
            _internal[ind]->set_joinValue(dMax);
            do_split(ind);
        }
    }

    void Diana::do_split(Size ind) const {
        std::vector<boost::shared_ptr<Node> > data = 
            _internal[ind]->data();
        Size n = data.size();

        Size ra;
        std::set<Size> splinter;
        std::set<Size> remaining;
        for(Size i=0; i<n; ++i) {
            Size id = data[i]->get_id();
            remaining.insert(id);
        }

        std::set<Size>::iterator it, it1;
        Real dMax = MIN_REAL;
        for(it = remaining.begin(); 
            it != remaining.end(); ++it) {
            Real dSum = 0.0;
            for(it1 = remaining.begin(); 
                it1 != remaining.end(); ++it1) {
                if(*it == *it1) {
                    continue;
                }
                dSum += _dm(*it, *it1);
            }
            if(dMax < dSum){
                dMax = dSum;
                ra = *it;
            }
        }
        splinter.insert(ra);
        remaining.erase(ra);

        bool bChanged = true;
        while(bChanged) {
            bChanged = false;
            for(it = remaining.begin(); 
                it != remaining.end(); ++it) {
                Real d1 = 0.0;
                for(it1 = splinter.begin(); 
                    it1 != splinter.end(); ++it1) {
                    d1 += _dm(*it, *it1);
                }
                d1 /= splinter.size();

                Real d2 = 0.0;
                for(it1 = remaining.begin(); 
                    it1 != remaining.end(); ++it1) {
                    if(*it == *it1) {
                        continue;
                    }
                    d2 += _dm(*it, *it1);
                }
                if(remaining.size() > 1) {
                    d2 /= (remaining.size()-1.0);
                }

                if(d1 < d2) {
                    bChanged = true;
                    splinter.insert(*it);
                    remaining.erase(it);
                    break;
                }
            }
        }

        _unsplitClusters.erase(ind); 
        _internal[ind]->clear();
        create_cluster(splinter, ind);
        create_cluster(remaining, ind);
    }

    void Diana::create_cluster(const std::set<Size> ele, 
        Size ind) const {
        std::set<Size>::iterator it;
        Real dMax;
        if(ele.size() > 1) {
            boost::shared_ptr<InternalNode> pin(new 
               InternalNode(0, _internal[ind]));
            _internal[ind]->add(pin);
            for(it = ele.begin(); it != ele.end(); ++it) {
                pin->add(_leaf[*it]);
            }

            it = _clusterID.end();
            --it;
            Size id = *it;
            _clusterID.erase(it);

            _internal.insert(std::pair<Size, 
                boost::shared_ptr<InternalNode> >(id, pin));
            _unsplitClusters.insert(id);

            dMax = MIN_REAL;
            std::vector<Size> nvTemp(ele.begin(), ele.end()); 
            for(Size i=0; i<nvTemp.size(); ++i) {
                for(Size j=i+1; j<nvTemp.size(); ++j) {
                    if(dMax < _dm(nvTemp[i], nvTemp[j])) {
                        dMax = _dm(nvTemp[i], nvTemp[j]);
                    }
                }
            } 
            _clusterDiameter.insert(
                std::pair<Size, Real>(id,dMax));
        } else {
            it = ele.begin();
            _internal[ind]->add(_leaf[*it]);
        } 
    }

    void Diana::fetchResults() const {
        Size n = _ds->size();
        HClustering hc(_internal[2*n-2]);
        _results.insert("hc", hc);
    }
}
