// cl/algorithms/fsc.cpp
#include<cl/algorithms/fsc.hpp>
#include<cl/clusters/pclustering.hpp>
#include<cl/errors.hpp>
#include<iostream> 
#include<boost/random.hpp>

namespace ClusLib {

    void FSC::performClustering() const { 
        initialization(); 
        iteration(); 
    }

    void FSC::setupArguments() { 
        Algorithm::setupArguments();
        ASSERT(_ds->is_numeric(), "dataset is not numeric");

        _epsilon = boost::any_cast<Real>(
            _arguments.get("epsilon")); 
        ASSERT(_epsilon>=0, "invalid epsilon");

        _threshold = boost::any_cast<Real>(
            _arguments.get("threshold"));
        ASSERT(_threshold>EPSILON, "invalid threshold");

        _alpha = boost::any_cast<Real>(_arguments.get("alpha"));
        ASSERT(_alpha>1, "invalid alpha");

        _numclust = boost::any_cast<Size>(
            _arguments.get("numclust")); 
        ASSERT(_numclust>=2 && _numclust<=_ds->size(), 
            "invalid numclust");

        _maxiter = boost::any_cast<Size>(
            _arguments.get("maxiter")); 
        ASSERT(_maxiter>0, "invalide maxiter");

        _seed = boost::any_cast<Size>(
            _arguments.get("seed")); 
        ASSERT(_seed>0, "invalide seed");
    }

    void FSC::fetchResults() const {
        PClustering pc;
        for(Size i=0;i<_clusters.size();++i){
            pc.add(_clusters[i]);
        }
        _results.CM = _CM;
        _results.insert("pc", boost::any(pc));
        _results.insert("numiter", boost::any(_numiter));
        _results.insert("dObj", boost::any(_dObj));
    }

    void FSC::iteration() const {
        Real dObjPre;

        updateWeight();
        updateCenter();
        _numiter = 1;
        while(true) { 
            Integer s;
            Real dMin,dDist;
            for(Size i=0;i<_ds->size();++i) {
                dMin = MAX_REAL;
                for(Size k=0;k<_clusters.size();++k) { 
                    dDist = dist((*_ds)[i],_clusters[k]);
                    if (dMin > dDist) {
                        dMin = dDist;
                        s = k;
                    }
                }

                if (_CM[i] != s){
                    _clusters[_CM[i]]->erase((*_ds)[i]);
                    _clusters[s]->add((*_ds)[i]);
                    _CM[i] = s;
                }
            }
            
            updateWeight();
            updateCenter();

            dObjPre = _dObj;
            calculateObj();
            if(std::fabs(dObjPre - _dObj) < _threshold){
                break;
            }

            ++_numiter;
            if (_numiter > _maxiter){
                break;
            }
        }
    }

    void FSC::updateCenter() const {
        Real dTemp;
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size k=0;k<_clusters.size();++k){ 
            for(Size j=0;j<schema->size();++j){
                dTemp = 0.0;
                for(Size i=0; i<_clusters[k]->size();++i){
                    boost::shared_ptr<Record> rec = 
                        (*_clusters[k])[i];
                    dTemp += (*schema)[j]->get_c_val((*rec)[j]);
                } 
                (*schema)[j]->set_c_val(
                    (*_clusters[k]->center())[j], 
                    dTemp/_clusters[k]->size());
            }
        }
    }

    void FSC::updateWeight() const {
        Real dVar, dSum;
        boost::shared_ptr<Schema> schema = _ds->schema();
        boost::shared_ptr<Record> c, r;
        std::vector<Real> w(schema->size(),0.0);
        for(Size k=0;k<_clusters.size();++k){ 
            c = _clusters[k]->center();
            dSum = 0.0;
            for(Size j=0;j<schema->size();++j){
                dVar = 0.0;
                for(Size i=0; i<_clusters[k]->size();++i){
                    r = (*_clusters[k])[i];
                    dVar += std::pow((*schema)[j]->distance(
                                (*r)[j], (*c)[j]), 2.0);
                } 
                w[j] = std::pow(1/(dVar + _epsilon),1/(_alpha-1));
                dSum += w[j];
            }
            for(Size j=0;j<schema->size();++j) {
                _clusters[k]->w(j) = w[j] / dSum;
            }
        }
    }

    void FSC::initialization() const {
        Size numRecords = _ds->size();
        std::vector<Integer> index(numRecords,0);
        _CM.resize(numRecords);
        for(Size i=0;i<index.size();++i){
            index[i] = i;
        }

        boost::minstd_rand generator(_seed);
        for(Size i=0;i<_numclust;++i){
            boost::uniform_int<> uni_dist(0,numRecords-i-1);
            boost::variate_generator<boost::minstd_rand&, 
                boost::uniform_int<> > uni(generator,uni_dist); 
            Integer r = uni();
            boost::shared_ptr<Record> cr = 
                boost::shared_ptr<Record>(new Record(*(*_ds)[r]));
            boost::shared_ptr<SubspaceCluster> c = 
                boost::shared_ptr<SubspaceCluster>(
                new SubspaceCluster(cr)); 
            c->set_id(i);
            _clusters.push_back(c);
            index.erase(index.begin()+r);
        }

        Integer s;
        Real dMin,dDist;
        for(Size i=0;i<numRecords;++i){
            dMin = MAX_REAL;
            for(Size j=0;j<_numclust;++j){ 
                dDist = dist((*_ds)[i], _clusters[j]);
                if (dDist<dMin){
                    s = j;
                    dMin = dDist;
                }
            }
            _clusters[s]->add((*_ds)[i]); 
            _CM[i] = s;
        } 
    }

    Real FSC::dist(const boost::shared_ptr<Record> &x,
              const boost::shared_ptr<SubspaceCluster> &c
              ) const {
        Real dTemp = 0.0;
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size j=0;j<schema->size();++j){
            dTemp += std::pow(c->w(j), _alpha) * 
                std::pow((*schema)[j]->distance((*x)[j], 
                (*c->center())[j]), 2.0);
        }

        return dTemp;
    }

    void FSC::calculateObj() const {
        Real dTemp = 0.0;
        for(Size i=0; i<_ds->size(); ++i){
            dTemp += dist((*_ds)[i], _clusters[_CM[i]]);
        }
        
        _dObj = dTemp;
    }
}
