// cl/algorithms/gkmode.cpp
#include<cl/algorithms/gkmode.hpp>
#include<boost/random.hpp>
#include<cl/types.hpp>
#include<set>
#include<map>

namespace ClusLib {

    void GKmode::setupArguments() {
        Algorithm::setupArguments();
        ASSERT(_ds->is_categorical(), 
            "dataset is not categorical");

        _numclust = boost::any_cast<Size>(
            _arguments.get("numclust")); 
        ASSERT(_numclust>=2 && _numclust<=_ds->size(), 
           "invalid numclust");

        _numpop = boost::any_cast<Size>(
            _arguments.get("numpop")); 
        ASSERT(_numpop>0, "invalide numpop");

        _maxgen = boost::any_cast<Size>(
            _arguments.get("maxgen")); 
        ASSERT(_maxgen>0, "invalide maxgen");
        
        _c = boost::any_cast<Real>(_arguments.get("c")); 
        ASSERT(_c>0 && _c<3, "c must be in range (0, 3)");

        _cm = boost::any_cast<Real>(_arguments.get("cm")); 
        ASSERT(_cm>0, "cm must be positive");

        _Pm = boost::any_cast<Real>(_arguments.get("pm")); 
        ASSERT(_Pm>0, "pm must be positive");
    }

    void GKmode::performClustering() const {
        initialization();
        iteration();
    }

    void GKmode::fetchResults() const {
        Real dMin = MAX_REAL;
        Size s;
        for(Size i=0; i<_numpop; ++i) {
            if (dMin > _dvFit[i]) {
                dMin = _dvFit[i];
                s = i;
            }
        }

        std::vector<Size> CM(_ds->size());
        for(Size i=0; i<_ds->size(); ++i){
            CM[i] = _mP(s, i);
        } 
        _results.CM = CM;

        Real dVar;
        Size mode;
        PClustering pc;
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size k=0; k<_numclust; ++k) {
            boost::shared_ptr<Record> r(new Record(schema));
            for(Size j=0; j<_ds->num_attr(); ++j) { 
                if (get_mode(mode, dVar, s, k, j) ) {
                    (*schema)[j]->set_d_val((*r)[j], mode);
                }
            }
            boost::shared_ptr<CenterCluster> c(new 
                CenterCluster(r));
            pc.add(c);
        } 

        for(Size i=0; i<_ds->size(); ++i){
            pc[CM[i]]->add( (*_ds)[i]);
        } 

        _results.insert("pc", boost::any(pc));
    }

    void GKmode::initialization() const {
        _mP.resize(_numpop, _ds->size());
        for(Size i=0; i<_numpop; ++i) {
            for(Size j=0; j<_ds->size(); ++j) {
                _mP(i,j) = Null<Size>();
            }
        }

        std::vector<Size> nvTemp(_numclust, 0);
        boost::minstd_rand generator(42u);
        boost::uniform_int<> uni_dist(0, _numclust-1);
        boost::variate_generator<boost::minstd_rand&, 
            boost::uniform_int<> > uni(generator,uni_dist); 
        for(Size i=0;i<_numpop;++i){
            generateRN(nvTemp, i);
            for(Size j=0; j<_numclust; ++j) {
                _mP(i, nvTemp[j]) = j;
            }
            for(Size j=0;j<_ds->size();++j) {
                if (_mP(i,j) == Null<Size>()) { 
                    _mP(i,j) = uni();
                }
            }
        }

        _dvFit.resize(_numpop, 0.0);
        _bvLegal.resize(_numpop, true);
    }

    void GKmode::iteration() const {
        Size g = 0;
        while (g < _maxgen) {
            ++g;
            Real Lmax = MIN_REAL;
            for(Size i=0; i<_numpop; ++i) {
                calculateLoss(i);
                if( _bvLegal[i] && _dvFit[i]>Lmax) {
                    Lmax = _dvFit[i];
                }
            }

            Real dTemp = 0.0;
            for(Size i=0; i<_numpop; ++i) {
                if( _bvLegal[i]) {
                    _dvFit[i] = _c*Lmax - _dvFit[i];
                } else {
                    _dvFit[i] = calculateRatio(i)*(_c-1)*Lmax;
                }
                dTemp += _dvFit[i];
            }

            for(Size i=0; i<_numpop; ++i) { 
                _dvFit[i] /= dTemp;
            }

            selection(g);
            mutation(g);

            for(Size i=0; i<_numpop; ++i) { 
                kmode(i);
            }
        }

        for(Size i=0; i<_numpop; ++i) { 
            calculateLoss(i);
        }
    }

    void GKmode::selection(Size g) const {
        boost::minstd_rand generator(
            static_cast<unsigned int>(g+1));
        boost::uniform_real<> uni_dist(0, 1);
        boost::variate_generator<boost::minstd_rand&, 
            boost::uniform_real<> > uni(generator,uni_dist); 
        Real dRand, dTemp;
        Size s;
        ublas::matrix<Size> mP = _mP;
        for(Size i=0;i<_numpop;++i){
            dTemp = _dvFit[0];
            dRand = uni();
            s = 0;
            while(dTemp < dRand) {
                dTemp += _dvFit[++s];
            }
            for(Size j=0; j<_ds->size(); ++j) {
                _mP(i, j) = mP(s, j);
            }
        } 
    }

    void GKmode::mutation(Size g) const { 
        boost::minstd_rand generator(
            static_cast<unsigned int>(g+1));
        boost::uniform_real<> uni_dist(0, 1);
        boost::variate_generator<boost::minstd_rand&, 
            boost::uniform_real<> > uni(generator,uni_dist); 
        Real dRand, dMax, dVar, dTemp; 
        Size mode; 
        bool isLegal;
        std::map<Size, Size> mTemp;
        std::vector<Real> dvProb(_numclust, 0.0);
        for(Size i=0; i<_numpop; ++i) {
            mTemp.clear();
            for(Size j=0; j<_ds->size(); ++j) {
                dRand = uni();
                if (dRand >= _Pm ) {
                    continue;
                }

                dMax = MIN_REAL;
                for(Size k=0; k<_numclust; ++k) {
                    dvProb[k] = 0.0;
                    for(Size d=0; d<_ds->num_attr(); ++d) {
                        isLegal = get_mode(mode, dVar, i, k, d);
                        if(!isLegal) {
                            break;
                        }
                        dvProb[k] += dVar;
                    }

                    if(dvProb[k] >dMax) {
                        dMax = dvProb[k];
                    } 
                }

                dTemp = 0.0;
                for(Size k=0; k<_numclust; ++k) {
                    dvProb[k] = _cm*dMax - dvProb[k];
                    dTemp += dvProb[k];
                }
                for(Size k=0; k<_numclust; ++k) {
                    dvProb[k] /= dTemp;
                }

                dRand = uni();
                dTemp = dvProb[0];
                Size k = 0;
                while( dTemp < dRand) {
                    dTemp += dvProb[++k];
                }
                mTemp.insert(std::pair<Size, Size>(j, k));
            }

            std::map<Size, Size>::iterator it;
            for(it = mTemp.begin(); it!=mTemp.end(); ++it) {
                _mP(i, it->first) = it->second;
            }
        }

    }

    void GKmode::kmode(Size ind) const {
        ublas::matrix<Size> 
            nmMode(_numclust, _ds->num_attr());

        bool isLegal;
        Real dVar;
        for(Size k=0; k<_numclust; ++k) {
            for(Size j=0; j<_ds->num_attr(); ++j) { 
                isLegal = get_mode(nmMode(k,j), dVar, ind, k, j);
                if(!isLegal) {
                    return;
                }
            }
        } 

        Real dDist;
        Real dMin; 
        Size h;
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size i=0; i<_ds->size(); ++i) {
            dMin = MAX_REAL;
            for(Size k=0; k<_numclust; ++k) {
                dDist = 0.0;
                for(Size j=0; j<_ds->num_attr(); ++j) {
                    if(nmMode(k, j) != (*schema)[j]->get_d_val(
                                (*(*_ds)[i])[j])) {
                        dDist += 1;
                    }
                }
                if(dMin > dDist) {
                    dMin = dDist; 
                    h = k;
                }
            } 
            _mP(ind, i) = h; 
        }
    }

    void GKmode::calculateLoss(Size ind) const {
        Real dLoss = 0.0; 
        Size mode;
        Real dVar;
        bool isLegal;
        for(Size k=0; k<_numclust; ++k){ 
            for(Size j=0; j<_ds->num_attr(); ++j) {
                isLegal = get_mode(mode, dVar, ind, k, j);
                dLoss += dVar;
                if(!isLegal) {
                    break;
                }
            }
        }

        _bvLegal[ind] = isLegal;
        if(isLegal) {
            _dvFit[ind] = dLoss;
        } else {
            _dvFit[ind] = MAX_REAL;
        }
    }

    bool GKmode::get_mode(Size &mode, Real &var, 
            Size ind, Size k, Size j) const {
        boost::shared_ptr<Schema> schema = _ds->schema();
        Size nValues = (*schema)[j]->cast_to_d().num_values();
        std::vector<Size> nvFreq(nValues,0);

        Size val;
        Size nCount = 0;
        for(Size i=0; i<_ds->size(); ++i) { 
            if (_mP(ind, i) == k) {
                val = (*schema)[j]->get_d_val( (*_ds)(i, j) );
                ++nvFreq[val];
                ++nCount;
            }
        }
        val = 0;
        for(Size i=0; i<nValues; ++i){
            if(val < nvFreq[i]) {
                val = nvFreq[i];
                mode = i;
            }
        }

        var = nCount - nvFreq[mode];

        if (nCount == 0) {
            return false;
        } else {
            return true;
        }
    }

    Real GKmode::calculateRatio(Size ind) const {
        std::set<Size> vals;
        for(Size i=0; i<_ds->size(); ++i) {
            vals.insert(_mP(ind, i));
        }

        return (Real)vals.size() / _numclust;
    }

    void GKmode::generateRN(std::vector<Size>& nv,
            Size s) const {
        boost::minstd_rand generator(
            static_cast<unsigned int>(s+1));
        for(Size i=0;i<_numclust;++i){
            boost::uniform_int<> uni_dist(0, _ds->size()-i-1);
            boost::variate_generator<boost::minstd_rand&, 
                boost::uniform_int<> > uni(generator,uni_dist); 
            nv[i] = uni();
        }
    }
}
