// cl/algorithms/gmc.cpp
#include<cl/algorithms/gmc.hpp>
#include<cl/clusters/pclustering.hpp>
#include<cl/distances/mahalanobisdistance.hpp>
#include<boost/random.hpp> 
#include<cmath>

namespace ClusLib {

    void GMC::setupArguments() {
        Algorithm::setupArguments();
        ASSERT(_ds->is_numeric(), "dataset is not numeric");

        _seed = boost::any_cast<Size>(
            _arguments.get("seed")); 
        ASSERT(_seed>0, "seed must be a positive number");

        _numclust = boost::any_cast<Size>(
            _arguments.get("numclust")); 
        ASSERT(_numclust>=2 && _numclust<=_ds->size(), 
            "invalid numclust");

        _maxiter = boost::any_cast<Size>(
            _arguments.get("maxiter")); 
        ASSERT(_maxiter>0, "invalide maxiter");

        _threshold = boost::any_cast<Real>(
            _arguments.get("threshold")); 
        ASSERT(_threshold>0, "threshold must be positive");

        _epsilon = boost::any_cast<Real>(
            _arguments.get("epsilon")); 
        ASSERT(_epsilon>=0, "epsilon must be nonnegative");
    }

    void GMC::performClustering() const {
        initialization();
        iteration();
    }

    void GMC::fetchResults() const {
        for(Size i=0; i<_ds->size(); ++i) {
            Real dMax = MIN_REAL;
            Integer k;
            for(Size j=0; j<_numclust; ++j) {
                if(dMax < _post(i,j)){
                    dMax = _post(i,j);
                    k = j;
                }
            }
            _CM[i] = k;
            _clusters[k]->add((*_ds)[i]);
        }

        PClustering pc;
        for(Size i=0;i<_clusters.size();++i){
            pc.add(_clusters[i]);
        }
        _results.insert("pc", boost::any(pc));
        _results.CM = _CM;
        _results.insert("converged", boost::any(_converged));
        _results.insert("p", boost::any(_p));
        _results.insert("numiter", boost::any(_numiter));
        _results.insert("ll", boost::any(_ll));
        _results.insert("mu", boost::any(_mu));
    }

    void GMC::iteration() const {
        Real pre_ll = MIN_REAL;
        _converged = false;
        _numiter = 0;
        for(Size iter=0; iter<_maxiter; ++iter) {
            estep();
            
            Real dTemp = _ll - pre_ll;
            if( dTemp >=0 && dTemp < _threshold) {
                _converged = true;
                break;
            }
            pre_ll = _ll;

            mstep();
            ++_numiter;
        }
    } 

    void GMC::estep() const { 
        Size numRecords = _ds->size();
        Size numAttr = _ds->num_attr();

        ublas::matrix<Real> log_ll(numRecords, _numclust);
        ublas::triangular_matrix<Real> L(numAttr, numAttr);

        for(Size j=0; j<_numclust; ++j) {
            Size k = chol(_sigma[j], L);
            ASSERT(k==0, "invalid covariance matrix " << j);

            Real logDet = 0.0;
            for(Size i=0; i<numAttr; ++i) {
                logDet += 2.0 * std::log(L(i,i));
            }

            MahalanobisDistance md(_sigma[j]);
            for(Size i=0; i<numRecords; ++i) {
                Real dist = md((*_ds)[i], _clusters[j]->center());
                log_ll(i,j) = 
                    - 0.5 * dist * dist
                    - 0.5 * logDet + std::log(_p[j]) 
                    - 0.5 * numAttr * std::log(2*3.141592653589);
            }
        }

        ublas::vector<Real> mlog_ll(numRecords);
        ublas::vector<Real> ones(_numclust);
        for(Size i=0; i<_numclust; ++i) {
            ones(i) = 1.0 / _numclust;
        }

        ublas::axpy_prod(log_ll, ones, mlog_ll, true);
        Real temp = 0.0;
        Real density;
        _ll = 0.0;
        for(Size i=0; i<numRecords; ++i) {
            density = 0.0;
            for(Size j=0; j<_numclust; ++j) {
                _post(i,j) = std::exp(log_ll(i,j)-mlog_ll(i));
                density += _post(i,j);
            }
            for(Size j=0; j<_numclust; ++j) {
                _post(i,j) /= density;
            }
            _ll += std::log(density) + mlog_ll(i);
        }
    }

    void GMC::mstep() const {
        boost::shared_ptr<Schema> schema = _ds->schema();
        Size numRecords = _ds->size();
        Size numAttr = _ds->num_attr();
        Real psum = 0.0;
        for(Size j=0; j<_numclust; ++j) {
            _p[j] = ublas::sum( ublas::column(_post, j));
            psum += _p[j];
        }

        ublas::matrix<Real> centered(numRecords, numAttr);
        for(Size k=0; k<_numclust; ++k) {
            for(Size j=0; j<numAttr; ++j) {
                _mu(k, j) = ublas::inner_prod(
                       ublas::column(_post, k),
                       ublas::column(_data, j) ) / _p[k];
                (*schema)[j]->set_c_val(
                    (*_clusters[k]->center())[j], _mu(k, j));
            }

            for(Size i=0; i<numRecords; ++i) {
                for(Size j=0; j<numAttr; ++j) {
                    centered(i,j) = std::sqrt(_post(i, k)) * 
                        (_data(i, j) - _mu(k, j));
                }
            }
            for(Size i=0; i<numAttr; ++i) {
                for(Size j=0; j<=i; ++j) { 
                    _sigma[k](i, j) =
                        ublas::inner_prod(
                            ublas::column(centered,i),
                            ublas::column(centered,j))/_p[k];
                }
                _sigma[k](i,i) += _epsilon;
            } 
        }

        for(Size j=0; j<_numclust; ++j) {
            _p[j] /= psum;
        }
    }

    void GMC::initialization() const {
        Size numRecords = _ds->size();
        Size numAttr = _ds->num_attr();

        _mu.resize(_numclust, numAttr);
        _sigma.resize(_numclust);
        _p.resize(_numclust);
        for(Size i=0; i<_numclust; ++i) {
            _sigma[i].resize(numAttr);
            _p[i] = 1.0 / _numclust;
        }
        _data.resize(numRecords, numAttr);
        _post.resize(numRecords, _numclust);

        std::vector<Real> mean(numAttr,0.0);
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size i=0; i<numRecords; ++i) {
            for(Size j=0; j<numAttr; ++j) {
                Real val = (*schema)[j]->get_c_val((*_ds)(i,j));
                _data(i,j) = val;
                mean[j] += val; 
            }
        }

        for(Size j=0; j<numAttr; ++j) {
            mean[j] /= numRecords; 
        }

        for(Size i=0; i<numAttr; ++i) {
            for(Size j=0; j<i; ++j) { 
                _sigma[0](i, j ) = 0.0;
            }
            _sigma[0](i, i) = ( ublas::inner_prod( 
                ublas::column(_data, i), 
                ublas::column(_data, i)) - 
                numRecords*mean[i]*mean[i]) 
                / (numRecords - 1.0) + _epsilon;
        } 

        for(Size i=1; i<_numclust; ++i) {
            _sigma[i] = _sigma[0];
        }

        _CM.resize(numRecords);
        std::vector<Integer> index(numRecords,0);
        for(Size i=0;i<index.size();++i){
            index[i] = i;
        }

        boost::minstd_rand generator(
            static_cast<unsigned int>(_seed));
        for(Size i=0;i<_numclust;++i){
            boost::uniform_int<> uni_dist(0,numRecords-i-1);
            boost::variate_generator<boost::minstd_rand&, 
                boost::uniform_int<> > uni(generator,uni_dist); 
            Integer r = uni();
            for(Size j=0; j<numAttr; ++j) {
                _mu(i, j) = _data(r,j);
            }

            boost::shared_ptr<Record> rec = (*_ds)[r];
            boost::shared_ptr<Record> cr = boost::shared_ptr
                <Record>(new Record(*rec));
            boost::shared_ptr<CenterCluster> c = boost::shared_ptr
                <CenterCluster>(new CenterCluster(cr)); 
            c->set_id(i);
            _clusters.push_back(c);
            index.erase(index.begin()+r);
        }
    } 
}
