// cl/algorithms/kmean.cpp
#include<cl/algorithms/kmean.hpp>
#include<cl/errors.hpp>
#include<iostream> 
#include<boost/random.hpp>

namespace ClusLib {

    void Kmean::performClustering() const { 
        initialization(); 
        iteration(); 
    }

    void Kmean::setupArguments() { 
        Algorithm::setupArguments();
        ASSERT(_ds->is_numeric(), "dataset is not numeric");

        _distance = _arguments.distance;
        ASSERT(_distance, "distance is null");

        _numclust = boost::any_cast<Size>(
            _arguments.get("numclust")); 
        ASSERT(_numclust>=2 && _numclust<=_ds->size(), 
            "invalid numclust");

        _maxiter = boost::any_cast<Size>(
            _arguments.get("maxiter")); 
        ASSERT(_maxiter>0, "invalide maxiter");

        _seed = boost::any_cast<Size>(
            _arguments.get("seed")); 
        ASSERT(_seed>0, "invalide seed");
    }

    void Kmean::fetchResults() const {
        PClustering pc;
        for(Size i=0;i<_clusters.size();++i){
            pc.add(_clusters[i]);
        }
        _results.CM = _CM;
        _results.insert("pc", boost::any(pc));

        _error = 0.0;
        for(Size i=0;i<_ds->size();++i) {
            _error += (*_distance)((*_ds)[i], 
                _clusters[_CM[i]]->center());
        }
        _results.insert("error", boost::any(_error));
        _results.insert("numiter", boost::any(_numiter));
    }

    void Kmean::iteration() const {
        bool bChanged = true;

        updateCenter();
        _numiter = 1;
        while(bChanged) { 
            bChanged = false;
            Size s;
            Real dMin,dDist;
            for(Size i=0;i<_ds->size();++i) {
                dMin = MAX_REAL;
                for(Size k=0;k<_clusters.size();++k) { 
                    dDist = (*_distance)((*_ds)[i], 
                        _clusters[k]->center());
                    if (dMin > dDist) {
                        dMin = dDist;
                        s = k;
                    }
                }

                if (_CM[i] != s){
                    _clusters[_CM[i]]->erase((*_ds)[i]);
                    _clusters[s]->add((*_ds)[i]);
                    _CM[i] = s;
                    bChanged = true;
                }
            }

            updateCenter();
            ++_numiter;
            if (_numiter > _maxiter){
                break;
            }
        }
    }

    void Kmean::updateCenter() const {
        Real dTemp;
        boost::shared_ptr<Schema> schema = _ds->schema();
        for(Size k=0;k<_clusters.size();++k){ 
            for(Size j=0;j<schema->size();++j){
                dTemp = 0.0;
                for(Size i=0; i<_clusters[k]->size();++i){
                    boost::shared_ptr<Record> rec = 
                        (*_clusters[k])[i];
                    dTemp += (*schema)[j]->get_c_val((*rec)[j]);
                } 
                (*schema)[j]->set_c_val(
                    (*_clusters[k]->center())[j],
                    dTemp/_clusters[k]->size());
            }
        }
    }

    void Kmean::initialization() const {
        Size numRecords = _ds->size();
        std::vector<Size> index(numRecords,0);
        _CM.resize(numRecords);
        for(Size i=0;i<index.size();++i){
            index[i] = i;
        }

        boost::minstd_rand generator(_seed);
        for(Size i=0;i<_numclust;++i){
            boost::uniform_int<> uni_dist(0,numRecords-i-1);
            boost::variate_generator<boost::minstd_rand&, 
                boost::uniform_int<> > uni(generator,uni_dist); 
            Size r = uni();
            boost::shared_ptr<Record> cr = boost::shared_ptr
                <Record>(new Record(*(*_ds)[r]));
            boost::shared_ptr<CenterCluster> c = boost::shared_ptr
                <CenterCluster>(new CenterCluster(cr)); 
            c->set_id(i);
            _clusters.push_back(c);
            index.erase(index.begin()+r);
        }

        Size s;
        Real dMin,dDist;
        for(Size i=0;i<numRecords;++i){
            dMin = MAX_REAL;
            for(Size j=0;j<_numclust;++j){ 
                dDist = (*_distance)((*_ds)[i], 
                    _clusters[j]->center());
                if (dDist<dMin){
                    s = j;
                    dMin = dDist;
                }
            }
            _clusters[s]->add((*_ds)[i]); 
            _CM[i] = s;
        } 
    } 
}
