// cl/utilities/datasetgenerator.cpp
#include<cl/utilities/datasetgenerator.hpp>
#include<sstream>

namespace ClusLib {

    DatasetGenerator::DatasetGenerator(ublas::matrix<Real> mu,
            std::vector<ublas::symmetric_matrix<Real> > sigma,
            std::vector<Size> records,
            Size seed): _mu(mu), _sigma(sigma), _records(records), 
        _seed(seed), _generator(boost::minstd_rand(seed), 
            boost::normal_distribution<>() )
    {
        ASSERT(_mu.size1() > 0 && _mu.size2()>0, 
            "empty input");
        ASSERT(_mu.size1()==records.size() &&
               _mu.size1()==sigma.size(),
            "number of clusters is not consistent");
        ASSERT(_mu.size2()==sigma[0].size1(),
            "number of attributes is not consistent");
        ASSERT(_seed>0, "seed must be postive");

        Size N = 0;
        for(Size i=0; i<_records.size(); ++i) {
            N += _records[i];
        }
        _data.resize(N, _mu.size2());
    }

    void DatasetGenerator::generate(Size ind) {
        ublas::triangular_matrix<Real> 
            T(_mu.size2(), _mu.size2()); 
        Size k = chol(_sigma[ind], T);
        ASSERT(k==0, "can not decompose sigma "<<ind);

        Size nStart = 0;
        for(Size i=0; i<ind; ++i) {
            nStart += _records[i];
        }

        ublas::vector<Real> v(_mu.size2());
        ublas::vector<Real> w(_mu.size2());
        for(Size i=0; i<_records[ind]; ++i) {
            for(Size j=0; j<v.size(); ++j) {
                v(j) = _generator();
            }
            ublas::axpy_prod(T, v, w);
            for(Size j=0; j<v.size(); ++j) {
                _data(nStart+i, j) = w(j) + _mu(ind, j);
            }
        }
    }

    void DatasetGenerator::fill(boost::shared_ptr<Dataset> &ds) {
        for(Size i=0; i<_records.size(); ++i) {
            generate(i);
        }

        boost::shared_ptr<Schema> schema(new Schema());
        boost::shared_ptr<DAttrInfo> labelInfo(
            new DAttrInfo("Label"));
        boost::shared_ptr<DAttrInfo> idInfo(
            new DAttrInfo("Identifier"));
        schema->labelInfo() = labelInfo;
        schema->idInfo() = idInfo;
        std::stringstream ss;
        for(Size s =0; s<_data.size2(); ++s) {
            ss.str("");
            ss<<"Attribute "<<s+1;
            boost::shared_ptr<CAttrInfo> cai(
                new CAttrInfo(ss.str()));
            schema->add(cai);
        }

        ds = boost::shared_ptr<Dataset>(new Dataset(schema));
        Size nCount = 0;
        for(Size i=0; i<_records.size(); ++i) {
            ss.str("");
            ss<<i+1;
            std::string label = ss.str();
            for(Size j=0; j<_records[i]; ++j) {
                ss.str("");
                ss<<nCount;
                std::string id = ss.str();
                boost::shared_ptr<Record> r(new Record(schema));
                schema->set_id(r, id);
                schema->set_label(r, label);

                for(Size s=0; s<_data.size2(); ++s) {
                    (*schema)[s]->set_c_val((*r)[s], 
                        _data(nCount, s));
                }
                ds->add(r);
                ++nCount;
            }
        }
    }
}
