// examples/cmean/cmean.cpp
#include<cl/cluslib.hpp>

#include<boost/timer.hpp>
#include<boost/program_options.hpp>
#include<iostream>
#include<fstream>
#include<iomanip>

using namespace ClusLib;
using namespace boost::program_options;

int main(int ac, char* av[]){
    try{
        options_description desc("Allowed options");
        desc.add_options()
            ("help", "produce help message")
            ("datafile", value<std::string>(), "the data file")
            ("k", value<Size>()->default_value(3), 
             "number of clusters")
            ("seed", value<Size>()->default_value(1), 
             "seed used to choose random initial centers")
            ("maxiter", value<Size>()->default_value(100), 
             "maximum number of iterations")
            ("numrun", value<Size>()->default_value(1), 
             "number of runs")
            ("epsilon",value<Real>()->default_value(1e-6, "1e-6"),
             "epsilon")
            ("alpha",value<Real>()->default_value(2.1,"2.1"),
             "alpha")
            ("threshold",
             value<Real>()->default_value(1e-12,"1e-12"),
             "Objective function tolerance");

        variables_map vm;        
        store(parse_command_line(ac, av, desc), vm);
        notify(vm);    

        if (vm.count("help") || ac==1) {
            std::cout << desc << "\n";
            return 1;
        }

        std::string datafile;
        if (vm.count("datafile")) {
            datafile = vm["datafile"].as<std::string>();
        } else {
            std::cout << "Please provide a data file\n";
            return 1;
        }

        Size numclust = vm["k"].as<Size>(); 
        Size maxiter = vm["maxiter"].as<Size>(); 
        Size numrun = vm["numrun"].as<Size>(); 
        Size seed = vm["seed"].as<Size>();
        Real alpha = vm["alpha"].as<Real>();
        Real epsilon = vm["epsilon"].as<Real>();
        Real threshold = vm["threshold"].as<Real>(); 

        DatasetReader reader(datafile);
        boost::shared_ptr<Dataset> ds; 
        reader.fill(ds);
        
        std::cout<<*ds<<std::endl;

        boost::timer t;
        t.restart();

        Results Res;
        Real avgiter = 0.0;
        Real avgerror = 0.0;
        Real dMin = MAX_REAL;
        Real error;
        for(Size i=1; i<=numrun; ++i) {
            Cmean ca;
            Arguments &Arg = ca.getArguments();
            Arg.ds = ds;
            Arg.insert("alpha", alpha);
            Arg.insert("epsilon", epsilon);
            Arg.insert("threshold", threshold);
            Arg.insert("numclust", numclust);
            Arg.insert("maxiter", maxiter);
            Arg.insert("seed", seed);
            if (numrun == 1) {
                Arg.additional["seed"] = seed;
            } else {
                Arg.additional["seed"] = i;
            }

            ca.clusterize();

            const Results &tmp = ca.getResults();
            avgiter += boost::any_cast<Size>(tmp.get("numiter"));
            error = boost::any_cast<Real>(tmp.get("dObj"));
            avgerror += error;  
            if (error < dMin) {
                dMin = error;
                Res = tmp;
            }
        }
        avgiter /= numrun;
        avgerror /= numrun;

        double seconds = t.elapsed();
        std::cout<<"completed in "<<seconds<<" seconds"
            <<std::endl;

        PClustering pc = 
            boost::any_cast<PClustering>(Res.get("pc"));
        std::cout<<pc<<std::endl;
        Size numiter = boost::any_cast<Size>(Res.get("numiter"));
        error = boost::any_cast<Real>(Res.get("dObj"));

        std::cout<<"\nNumber of run: "<<numrun<<std::endl;
        std::cout<<"Average number of iterations: "
            <<avgiter<<std::endl;
        std::cout<<"Average error: "<<avgerror<<std::endl;
        std::cout<<"Number of iterations for the best case: "
            <<numiter<<std::endl;
        std::cout<<"Best error: "<<error<<std::endl;

        boost::numeric::ublas::matrix<Real> fcm =
            boost::any_cast<boost::numeric::ublas::matrix<Real> >(
            Res.get("fcm"));
        std::cout<<"\nFuzzy cluster memberships of "
            <<"the first 5 records:"<<std::endl;
        for(Size i=0;i<fcm.size1();++i){
            std::cout<<"Record "<<i;
            for(Size j=0;j<fcm.size2();++j){
                std::cout<<", "<<fcm(i,j);
            }
            std::cout<<std::endl;
            if (i==4 && fcm.size1()>4) {
                std::cout<<"..."<<std::endl;
                break;
            }
        }

        return 0;
    } catch (std::exception& e) {
        std::cout<<e.what()<<std::endl;
        return 1;
    } catch (...){
        std::cout<<"unknown error"<<std::endl;
        return 2;
    }
}
