// examples/gmc/gmc.cpp
#include<cl/cluslib.hpp>

#include<boost/timer.hpp>
#include<boost/program_options.hpp>
#include<iostream>
#include<fstream>
#include<iomanip>

using namespace ClusLib;
using namespace boost::program_options;

int main(int ac, char* av[]){
    try{
        options_description desc("Allowed options");
        desc.add_options()
            ("help", "produce help message")
            ("datafile", value<std::string>(), "the data file")
            ("k", value<Size>()->default_value(3), 
             "number of clusters")
            ("seed", value<Size>()->default_value(1), 
             "seed used to choose random initial centers")
            ("maxiter", value<Size>()->default_value(100), 
             "maximum number of iterations")
            ("numrun", value<Size>()->default_value(1), 
             "number of runs")
            ("threshold", value<Real>()->default_value(1e-10),
             "Likelihood tolerance")
            ("epsilon", value<Real>()->default_value(0.0),
             "Regularization parameter");

        variables_map vm;        
        store(parse_command_line(ac, av, desc), vm);
        notify(vm);    

        if (vm.count("help") || ac==1) {
            std::cout << desc << "\n";
            return 1;
        }

        std::string datafile;
        if (vm.count("datafile")) {
            datafile = vm["datafile"].as<std::string>();
        } else {
            std::cout << "Please provide a data file\n";
            return 1;
        }

        Size numclust = vm["k"].as<Size>(); 
        Size maxiter = vm["maxiter"].as<Size>(); 
        Size numrun = vm["numrun"].as<Size>(); 
        Size seed = vm["seed"].as<Size>();
        Real threshold = vm["threshold"].as<Real>();
        Real epsilon = vm["epsilon"].as<Real>();

        if(numrun ==0 ){
            return 1;
        }

        DatasetReader reader(datafile);
        boost::shared_ptr<Dataset> ds; 
        reader.fill(ds);
        
        std::cout<<*ds<<std::endl;

        boost::timer t;
        t.restart();

        Results Res;
        Real avgiter = 0.0;
        Real avgll = 0.0; 
        Real maxll = MIN_REAL;
        for(Size i=1; i<=numrun; ++i) {
            GMC ca;
            Arguments &Arg = ca.getArguments();
            Arg.ds = ds;
            Arg.insert("numclust", numclust);
            Arg.insert("maxiter", maxiter);
            if(numrun == 1) {
                Arg.insert("seed", seed);
            } else {
                Arg.insert("seed", i);
            }
            Arg.insert("epsilon", epsilon);
            Arg.insert("threshold", threshold);

            ca.clusterize();
            const Results &tmp = ca.getResults(); 
            Real ll = boost::any_cast<Real>(tmp.get("ll"));
            avgll += ll;
            if (ll > maxll) {
                maxll = ll;
                Res = tmp;
            }
            avgiter += boost::any_cast<Size>(tmp.get("numiter"));
        }
        avgiter /= numrun;
        avgll /= numrun;

        double seconds = t.elapsed();
        std::cout<<"completed in "<<seconds<<" seconds"
            <<std::endl;

        PClustering pc = 
            boost::any_cast<PClustering>(Res.get("pc"));
        std::cout<<pc<<std::endl;

        pc.save("iris.txt"); 

        std::vector<Real> p = boost::any_cast<std::vector<Real> >(
                Res.get("p"));
        ublas::matrix<Real> mu = 
            boost::any_cast<ublas::matrix<Real> >(Res.get("mu"));
        std::cout<<"Component size: "<<std::endl;
        for(Size i=0; i<p.size(); ++i){
            std::cout<<"Cluster "<<i<<": "<<p[i]<<std::endl;
        }
        std::cout<<"\nCluster Center: "<<std::endl;
        for(Size i=0; i<p.size(); ++i){
            std::cout<<"Center "<<i<<" "
                <<ublas::row(mu,i)<<std::endl;
        }
        std::cout<<"\nNumber of runs: "<<numrun<<std::endl;
        std::cout<<"Average number of iterations: "
            <<avgiter<<std::endl;
        std::cout<<"Average likelihood: "
            <<avgll<<std::endl;
        std::cout<<"Best likelihood: "<<maxll<<std::endl;
        std::cout<<"Number of iterations for the best case: "
            <<boost::any_cast<Size>(Res.get("numiter"))
            <<std::endl;

        return 0;
    } catch (std::exception& e) {
        std::cout<<e.what()<<std::endl;
        return 1;
    } catch (...){
        std::cout<<"unknown error"<<std::endl;
        return 2;
    }
}
