// examples/kprototype/kprototype.cpp
#include<cl/cluslib.hpp>

#include<boost/timer.hpp>
#include<boost/program_options.hpp>
#include<iostream>
#include<fstream>
#include<iomanip>

using namespace ClusLib;
using namespace std;
using namespace boost::program_options;

int main(int ac, char* av[]){
    try{
        options_description desc("Allowed options");
        desc.add_options()
            ("help", "produce help message")
            ("datafile", value<string>(), "the data file")
            ("normalize", value<string>()->default_value("no"), 
             "normalize the data or not")
            ("k", value<Size>()->default_value(3), 
             "number of clusters")
            ("beta", value<Real>()->default_value(1), 
             "balance weight for distance")
            ("seed", value<Size>()->default_value(1), 
             "seed used to choose random initial centers")
            ("maxiter", value<Size>()->default_value(100), 
             "maximum number of iterations")
            ("numrun", value<Size>()->default_value(1), 
             "number of runs");

        variables_map vm;        
        store(parse_command_line(ac, av, desc), vm);
        notify(vm);    

        if (vm.count("help") || ac==1) {
            cout << desc << "\n";
            return 1;
        }

        string datafile;
        if (vm.count("datafile")) {
            datafile = vm["datafile"].as<string>();
        } else {
            cout << "Please provide a data file\n";
            return 1;
        }

        Real beta = vm["beta"].as<Real>();
        Size numclust = vm["k"].as<Size>(); 
        Size maxiter = vm["maxiter"].as<Size>(); 
        Size numrun = vm["numrun"].as<Size>(); 
        Size seed = vm["seed"].as<Size>();
        string normalize = vm["normalize"].as<string>();

        DatasetReader reader(datafile);
        boost::shared_ptr<Dataset> ds; 
        reader.fill(ds);

        if( normalize != "no" ) {
            boost::shared_ptr<Dataset> ods = ds;
            DatasetNormalizer dn(ods);
            dn.fill(ds);
        }

        std::cout<<*ds<<std::endl;

        boost::shared_ptr<Distance> dist(new MixedDistance(beta)); 

        boost::timer t;
        t.restart();
        
        Results Res;
        Real avgiter = 0.0;
        Real avgerror = 0.0;
        Real dMin = MAX_REAL;
        Real error;
        for(Size i=1; i<=numrun; ++i) {
            Kprototype ca;
            Arguments &Arg = ca.getArguments();
            Arg.ds = ds;
            Arg.distance = dist;
            Arg.insert("numclust", numclust);
            Arg.insert("maxiter", maxiter);
            Arg.insert("seed", seed);
            if (numrun == 1) {
                Arg.additional["seed"] = seed;
            } else {
                Arg.additional["seed"] = i;
            }

            ca.clusterize();

            const Results &tmp = ca.getResults();
            avgiter += boost::any_cast<Size>(tmp.get("numiter"));
            error = boost::any_cast<Real>(tmp.get("error"));
            avgerror += error;  
            if (error < dMin) {
                dMin = error;
                Res = tmp;
            }
        }
        avgiter /= numrun;
        avgerror /= numrun;

        double seconds = t.elapsed();
        std::cout<<"completed in "<<seconds<<" seconds"
            <<std::endl;

        PClustering pc = 
            boost::any_cast<PClustering>(Res.get("pc"));

        std::cout<<pc<<std::endl;
        std::cout<<"Number of runs: "<<numrun<<std::endl;
        std::cout<<"Average number of iterations: "
            <<avgiter<<std::endl;
        std::cout<<"Average error: "<<avgerror<<std::endl;
        std::cout<<"Best error: "<<dMin<<std::endl;

        return 0;
    } catch (std::exception& e) {
        std::cout<<e.what()<<std::endl;
        return 1;
    } catch (...){
        std::cout<<"unknown error"<<std::endl;
        return 2;
    }
}
