// examples/mpikmean/mpikmean.cpp
#include<iostream> 
#include<boost/random.hpp>
#include<cmath>
#include<mpikmean.hpp>
#include<boost/serialization/vector.hpp>

namespace ClusLib {

    void MPIKmean::performClustering() const { 
        initialization(); 
        iteration(); 
    }

    void MPIKmean::setupArguments() { 
        _numclust = boost::any_cast<Size>(
            _arguments.get("numclust")); 

        _maxiter = boost::any_cast<Size>(
            _arguments.get("maxiter")); 
        ASSERT(_maxiter>0, "invalide maxiter");

        _seed = boost::any_cast<Size>(
            _arguments.get("seed")); 
        ASSERT(_seed>0, "invalide seed"); 

        if(_world.rank() ==0) {
            Algorithm::setupArguments();
            ASSERT(_ds->is_numeric(), "dataset is not numeric");

            ASSERT(_numclust>=2 && _numclust<=_ds->size(), 
                "invalid numclust");
        }

    }

    void MPIKmean::fetchResults() const {
        std::vector<Real> error(1, 0.0), totalerror(1);
        for(Size i=0;i<_numObj;++i) {
            error[0] += dist(i,_CM[i]);
        } 

        reduce(_world, error, totalerror, vplus<Real>(), 0);

        if(_world.rank() == 0) { 
            boost::shared_ptr<Schema> schema = _ds->schema();
            PClustering pc;
            for(Size i=0;i<_numclust;++i){
                for(Size j=0; j<_numAttr; ++j) { 
                    (*schema)[j]->set_c_val(
                        (*_clusters[i]->center())[j], 
                        _centers[i*_numAttr+j]);
                }
                pc.add(_clusters[i]);
            }

            for(Size i=0; i<_CM.size(); ++i) {
                _clusters[_CM[i]]->add((*_ds)[i]);
            }

            _results.CM = _CM;
            _results.insert("pc", boost::any(pc));

            _error = totalerror[0];
            _results.insert("error", boost::any(_error));
            _results.insert("numiter", boost::any(_numiter));
        }
    }

    void MPIKmean::iteration() const {
        std::vector<Size> nChanged(1,1); 

        _numiter = 1;
        while(nChanged[0] > 0) { 
            nChanged[0] = 0;
            Size s;
            Real dMin,dDist;
            std::vector<Size> nChangedLocal(1,0);
            std::vector<Real> newCenters(_numclust*_numAttr,0.0);
            std::vector<Size> newSize(_numclust,0);

            for(Size i=0;i<_numObj;++i) {
                dMin = MAX_REAL;
                for(Size k=0;k<_numclust;++k) { 
                    dDist = dist(i, k);
                    if (dMin > dDist) {
                        dMin = dDist;
                        s = k;
                    }
                }

                for(Size j=0; j<_numAttr; ++j) {
                    newCenters[s*_numAttr+j] += 
                    		_data[i*_numAttr+j];
                }
                newSize[s] +=1;

                if (_CM[i] != s){
                    _CM[i] = s;
                    nChangedLocal[0]++;
                }
            }

            all_reduce(_world, nChangedLocal, nChanged, 
            		vplus<Size>());
            all_reduce(_world, newCenters, _centers, 
            		vplus<Real>()); 
            std::vector<Size> totalSize(_numclust,0);
            all_reduce(_world, newSize, totalSize, vplus<Size>()); 

            for(Size k=0; k<_numclust; ++k) {
                for(Size j=0; j<_numAttr; ++j) {
                    _centers[k*_numAttr+j] /= totalSize[k];
                }
            }

            ++_numiter;
            if (_numiter > _maxiter){
                break;
            }
        }

        if(_world.rank() > 0) {
            _world.send(0,0,_CM);
        } else {
            for(Size p=1; p<_world.size(); ++p) {
                std::vector<Size> msg;
                _world.recv(p,0,msg);
                for(Size j=0; j<msg.size(); ++j) {
                    _CM.push_back(msg[j]);
                }
            }
        }
    }

    void MPIKmean::initialization() const {
        Size numRecords; 
        Size rank = _world.rank();

        if (rank == 0) {
            numRecords = _ds->size(); 
            _numAttr = _ds->num_attr();
            _centers.resize(_numclust * _numAttr);

            std::vector<Integer> index(numRecords,0);
            for(Size i=0;i<index.size();++i){
                index[i] = i;
            }

            boost::shared_ptr<Schema> schema = _ds->schema();
            boost::minstd_rand generator(_seed);
            for(Size i=0;i<_numclust;++i){
                boost::uniform_int<> uni_dist(0,numRecords-i-1);
                boost::variate_generator<boost::minstd_rand&, 
                    boost::uniform_int<> > 
                        uni(generator,uni_dist); 
                Integer r = uni();
                boost::shared_ptr<Record> cr = boost::shared_ptr
                    <Record>(new Record(*(*_ds)[r]));
                boost::shared_ptr<CenterCluster> c = 
                    boost::shared_ptr<CenterCluster>(
                        new CenterCluster(cr)); 
                c->set_id(i);
                _clusters.push_back(c);
                for(Size j=0; j<_numAttr; ++j) {
                    _centers[i*_numAttr + j] = 
                        (*schema)[j]->get_c_val((*_ds)(r,j));
                }
                index.erase(index.begin()+r);
            } 

        } 
        
        boost::mpi::broadcast(_world, _centers, 0);
        boost::mpi::broadcast(_world, numRecords, 0);
        boost::mpi::broadcast(_world, _numAttr, 0);

        Size nDiv = numRecords / _world.size();
        Size nRem = numRecords % _world.size();

        if(rank == 0) { 
            boost::shared_ptr<Schema> schema = _ds->schema();
            _numObj = (nRem >0) ? nDiv+1: nDiv; 
            _data.resize(_numObj * _numAttr);
            _CM.resize(_numObj);
            for(Size i=0; i<_numObj; ++i) {
                for(Size j=0; j<_numAttr; ++j) {
                    _data[i*_numAttr +j] = 
                        (*schema)[j]->get_c_val((*_ds)(i, j));
                }
            }

            Size nCount = _numObj; 
            for(Size p=1; p<_world.size(); ++p) {
                Size s = (p< nRem) ? nDiv +1 : nDiv;
                std::vector<Real> dv(s*_numAttr);
                for(Size i=0; i<s; ++i) {
                    for(Size j=0; j<_numAttr; ++j) { 
                        dv[i*_numAttr+j] = 
                            (*schema)[j]->get_c_val(
                            		(*_ds)(i+nCount,j));
                    }
                }
                nCount += s;
                _world.send(p, 0, dv);
            }
        } else {
            _numObj = (rank < nRem) ? nDiv+1: nDiv; 
            _CM.resize(_numObj);
            _world.recv(0,0,_data);
        } 
    } 

    Real MPIKmean::dist(Size i, Size j) const {
        Real dDist = 0.0;
        for(Size h=0; h<_numAttr; ++h) {
            dDist += std::pow(_data[i*_numAttr + h] 
                    - _centers[j*_numAttr + h], 2.0);
        }
        return std::pow(dDist, 0.5);
    }
}
