// test-suite/distance.cpp
#include"distance.hpp"

#include<cl/datasets/dataset.hpp>
#include<cl/distances/euclideandistance.hpp>
#include<cl/distances/mahalanobisdistance.hpp>
#include<cl/distances/minkowskidistance.hpp>
#include<cl/utilities/datasetreader.hpp>
#include<sstream>
#include<fstream>

using namespace ClusLib;
using namespace boost::unit_test_framework;

void DistanceTest::testEuclidean() {
    BOOST_MESSAGE("Testing Euclidean"); 

    std::string fileName("../../Data/iris.data");
    DatasetReader reader(fileName);
    boost::shared_ptr<Dataset> ds;
    reader.fill(ds);

    EuclideanDistance ed;
    std::vector<Real> dist(ds->size());
    for(Size i=0;i<ds->size();++i){
        dist[i] = ed((*ds)[0],(*ds)[i]);
    }

    std::stringstream ss;
    for(Size i=0;i<dist.size();++i){
        ss<<"distance between records 0 and "<<(*ds)[i]->get_id()
            <<": "<<dist[i]<<'\n';
    }


    std::ofstream out;
    out.open("irisdist.txt"); 
    out<<ss.str();
    out.close();
}

void DistanceTest::testMahalanobis() {
    BOOST_MESSAGE("Testing Mahalanobis"); 

    std::string fileName("../../Data/bezdekIris.data");
    DatasetReader reader(fileName);
    boost::shared_ptr<Dataset> ds;
    reader.fill(ds);

    boost::shared_ptr<Schema> schema = ds->schema();
    ublas::matrix<Real> data;
    ublas::vector<Real> mu;

    data.resize(ds->size(), ds->num_attr());
    mu.resize(ds->num_attr());
    for(Size j=0; j<mu.size(); ++j) {
        mu(j) = 0.0;
    }
    for(Size i=0; i<ds->size(); ++i) {
        for(Size j=0; j<ds->num_attr(); ++j) {
            data(i,j) = (*schema)[j]->get_c_val((*ds)(i,j)); 
            mu(j) += data(i,j);
        }
    }

    ublas::symmetric_matrix<Real> cov;
    cov.resize(ds->num_attr(), ds->num_attr());
    for(Size i=0; i<ds->num_attr(); ++i) {
        for(Size j=0; j<=i; ++j) { 
            cov(i, j) = ( inner_prod( column(data, i), 
                column(data, j)) - mu(i)*mu(j) / ds->size()) 
                / (ds->size() - 1.0);
        }
    } 

    std::cout<<"cov: "<<std::endl;
    std::cout<<cov<<std::endl;

    boost::shared_ptr<Record> r(new Record(ds->schema()));
    for(Size i=0; i<r->size(); ++i) {
        (*schema)[i]->set_c_val((*r)[i], mu(i)/ds->size());
    }

    MahalanobisDistance md(cov);
    std::vector<Real> dist(ds->size());
    for(Size i=0;i<ds->size();++i){
        dist[i] = md(r,(*ds)[i]);
    }

    std::stringstream ss;
    for(Size i=0;i<dist.size();++i){
        ss<<"distance between records 0 and "<<(*ds)[i]->get_id()
            <<": "<<dist[i]<<'\n';
    }


    std::ofstream out;
    out.open("iris_mahal.txt"); 
    out<<ss.str();
    out.close();
}

test_suite* DistanceTest::suite() { 
    test_suite* suite = BOOST_TEST_SUITE("Testing Distances");

    suite->add(BOOST_TEST_CASE(&DistanceTest::testEuclidean));
    suite->add(BOOST_TEST_CASE(&DistanceTest::testMahalanobis));

    return suite;
}
